@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/child.c,v 1.16 2006/06/26 08:25:17 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/child.c,v 1.25 2007/08/01 04:25:40 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2006	PgPool Global Development Group
+ * Copyright (c) 2003-2007	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -27,6 +27,8 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <sys/un.h>
+#include <arpa/inet.h>
+#include <netdb.h>
 #ifdef HAVE_NETINET_TCP_H
 #include <netinet/tcp.h>
 #endif
@@ -45,6 +47,7 @@
 #endif
 
 #include "pool.h"
+#include "pool_ip.h"
 
 #ifdef NONE_BLOCK
 static void set_nonblock(int fd);
@@ -68,6 +71,11 @@ static int exit_request;
 
 static int idle;		/* non 0 means this child is in idle state */
 
+extern int myargc;
+extern char **myargv;
+
+char remote_ps_data[NI_MAXHOST];		/* used for set_ps_display */
+
 /*
 * child main loop
 */
@@ -83,9 +91,13 @@ void do_child(int unix_fd, int inet_fd)
 	int connections_count = 0;	/* used if child_max_connections > 0 */
 	int first_ready_for_query_received;		/* for master/slave mode */
 	int found;
+	char psbuf[NI_MAXHOST + 128];
 
 	pool_debug("I am %d", getpid());
 
+	/* Identify myself via ps */
+	init_ps_display("", "", "", "");
+
 	/* set up signal handlers */
 	signal(SIGALRM, SIG_DFL);
 	signal(SIGTERM, die);
@@ -165,6 +177,13 @@ void do_child(int unix_fd, int inet_fd)
 		idle = 0;
 		child_idle_sec = 0;
 
+		/* check backend timer is expired */
+		if (backend_timer_expired)
+		{
+			pool_backend_timer();
+			backend_timer_expired = 0;
+		}
+
 		/* disable timeout */
 		pool_disable_timeout();
 
@@ -211,6 +230,29 @@ void do_child(int unix_fd, int inet_fd)
 			goto retry_startup;
 		}
 
+		if (pool_config.enable_pool_hba)
+		{
+			/*
+			 * do client authentication.
+			 * Note that ClientAuthentication does not return if frontend
+			 * was rejected; it simply terminates this process.
+			 */
+			frontend->protoVersion = sp->major;
+			frontend->database = strdup(sp->database);
+			if (frontend->database == NULL)
+			{
+				pool_error("do_child: strdup failed: %s\n", strerror(errno));
+				exit(1);
+			}
+			frontend->username = strdup(sp->user);
+			if (frontend->username == NULL)
+			{
+				pool_error("do_child: strdup failed: %s\n", strerror(errno));
+				exit(1);
+			}
+			ClientAuthentication(frontend);
+		}
+
 		/*
 		 * Ok, negotiaton with frontend has been done. Let's go to the next step.
 		 */
@@ -224,7 +266,7 @@ void do_child(int unix_fd, int inet_fd)
 
 		/* look for existing connection */
 		found = 0;
-		backend = pool_get_cp(sp->user, sp->database, sp->major);
+		backend = pool_get_cp(sp->user, sp->database, sp->major, 1);
 
 		if (backend != NULL)
 		{
@@ -327,6 +369,12 @@ void do_child(int unix_fd, int inet_fd)
 
 		connected = 1;
 
+		/* show ps status */
+		sp = MASTER_CONNECTION(backend)->sp;
+		snprintf(psbuf, sizeof(psbuf), "%s %s %s idle",
+				 sp->user, sp->database, remote_ps_data);
+		set_ps_display(psbuf, false);
+
 		/* query process loop */
 		for (;;)
 		{
@@ -384,9 +432,13 @@ void do_child(int unix_fd, int inet_fd)
 					exit(1);
 					break;
 
-				/* fatal error occured. just exit myself... */
+				/*
+				 * kind mismatch fatal error occured.
+				 * notice that we need to detach secondary server
+				 * and just exit myself...
+				 */
 				case POOL_FATAL:
-					notice_backend_error(1);
+					notice_backend_error(0);
 					exit(1);
 					break;
 
@@ -479,9 +531,9 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 {
     fd_set	readmask;
     int fds;
+	int save_errno;
 
-	struct sockaddr addr;
-	socklen_t addrlen;
+	SockAddr saddr;
 	int fd = 0;
 	int afd;
 	int inet = 0;
@@ -492,7 +544,12 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	static int cnt;
 #endif
 	struct timeval *timeoutval;
-	struct timeval tv1, tv2, tmback;
+	struct timeval tv1, tv2, tmback = {0, 0};
+
+	char remote_host[NI_MAXHOST];
+	char remote_port[NI_MAXSERV];
+
+	set_ps_display("wait for connection request", false);
 
 	FD_ZERO(&readmask);
 	FD_SET(unix_fd, &readmask);
@@ -516,6 +573,14 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 
 	fds = select(Max(unix_fd, inet_fd)+1, &readmask, NULL, NULL, timeoutval);
 
+	save_errno = errno;
+	/* check backend timer is expired */
+	if (backend_timer_expired)
+	{
+		pool_backend_timer();
+		backend_timer_expired = 0;
+	}
+
 	/*
 	 * following code fragment computes remaining timeout val in a
 	 * portable way. Linux does this automazically but other platforms do not.
@@ -548,6 +613,8 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 #endif
 	}
 
+	errno = save_errno;
+
 	if (fds == -1)
 	{
 		if (errno == EAGAIN || errno == EINTR)
@@ -578,19 +645,29 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	 * Note that some SysV systems do not work here. For those
 	 * systems, we need some locking mechanism for the fd.
 	 */
-	addrlen = sizeof(addr);
+	memset(&saddr, 0, sizeof(saddr));
+	saddr.salen = sizeof(saddr.addr);
 
 #ifdef ACCEPT_PERFORMANCE
 	gettimeofday(&now1,0);
 #endif
-	afd = accept(fd, &addr, &addrlen);
+	afd = accept(fd, (struct sockaddr *)&saddr.addr, &saddr.salen);
+
+	save_errno = errno;
+	/* check backend timer is expired */
+	if (backend_timer_expired)
+	{
+		pool_backend_timer();
+		backend_timer_expired = 0;
+	}
+	errno = save_errno;
 	if (afd < 0)
 	{
 		/*
 		 * "Resource temporarily unavailable" (EAGAIN or EWOULDBLOCK)
-		 * can be silently ignored.
+		 * can be silently ignored. And EINTR can be ignored
 		 */
-		if (errno != EAGAIN && errno != EWOULDBLOCK)
+		if (errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
 			pool_error("accept() failed. reason: %s", strerror(errno));
 		return NULL;
 	}
@@ -603,8 +680,23 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 		pool_log("cnt: %d atime: %ld", cnt, atime);
 	}
 #endif
+
 	pool_debug("I am %d accept fd %d", getpid(), afd);
 
+	pool_getnameinfo_all(&saddr, remote_host, remote_port);
+	snprintf(remote_ps_data, sizeof(remote_ps_data),
+			 remote_port[0] == '\0' ? "%s" : "%s(%s)",
+			 remote_host, remote_port);
+
+	set_ps_display("accept connection", false);
+
+	/* log who is connecting */
+	if (pool_config.log_connections)
+	{
+		pool_log("connection received: host=%s%s%s",
+				 remote_host, remote_port[0] ? " port=" : "", remote_port);
+	}
+
 	/* set NODELAY and KEEPALIVE options if INET connection */
 	if (inet)
 	{
@@ -633,6 +725,12 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 		close(afd);
 		return NULL;
 	}
+
+	/* save ip addres for hba */
+	memcpy(&cp->raddr, &saddr, sizeof(SockAddr));
+	if (cp->raddr.addr.ss_family == 0)
+		cp->raddr.addr.ss_family = AF_UNIX;
+	
 	return cp;
 }
 
