@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.16 2007/02/01 15:31:59 yamaguti Exp $
+ * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.22 2007/08/10 03:57:44 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -21,11 +21,15 @@
 #include "pool.h"
 
 #include <sys/types.h>
+#include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <sys/un.h>
 #include <netdb.h>
 #include <arpa/inet.h>
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -46,6 +50,20 @@
 
 #include "version.h"
 
+#define CHECK_REQUEST \
+	do \
+	{ \
+		if (failover_request) \
+		{ \
+			failover(failover_signo); \
+			failover_request = 0; \
+		} \
+		if (sigchld_request) \
+		{ \
+			reaper(); \
+		} \
+	} while (0)
+
 #define PGPOOLMAXLITSENQUEUELENGTH 10000
 static void daemonize(void);
 static int read_pid_file(void);
@@ -54,6 +72,10 @@ static pid_t fork_a_child(int unix_fd, int inet_fd);
 static int create_unix_domain_socket(void);
 static int create_inet_domain_socket(const char *hostname);
 static void myexit(int code);
+static void failover(int sig);
+static void reaper(void);
+static int pool_pause(struct timeval *timeout);
+static void pool_sleep(unsigned int second);
 
 static RETSIGTYPE exit_handler(int sig);
 static RETSIGTYPE reap_handler(int sig);
@@ -86,6 +108,10 @@ static int stop_sig = SIGTERM;	/* stopping signal default value */
 static int switch_over_sig = SIGUSR1;	/* switch over signal default value */
 
 static int health_check_timer_expired;		/* non 0 if health check timer expired */
+static volatile sig_atomic_t failover_request = 0;
+static volatile sig_atomic_t sigchld_request = 0;
+static int pipe_fds[2]; /* for delivering signals */
+static volatile int failover_signo = 0;
 
 int myargc;
 char **myargv;
@@ -296,6 +322,7 @@ int main(int argc, char **argv)
 	}
 
 	/* set up signal handlers */
+	POOL_SETMASK(&BlockSig);
 	pool_signal(SIGTERM, exit_handler);
 	pool_signal(SIGINT, exit_handler);
 	pool_signal(SIGQUIT, exit_handler);
@@ -304,6 +331,13 @@ int main(int argc, char **argv)
 	pool_signal(SIGUSR2, failover_handler);
 	pool_signal(SIGHUP, exit_handler);
 
+	/* create pipe for delivering event */
+	if (pipe(pipe_fds) < 0)
+	{
+		pool_error("failed to create pipe");
+		myexit(1);
+	}
+
 	pool_log("pgpool successfully started");
 
 	/*
@@ -311,6 +345,8 @@ int main(int argc, char **argv)
 	 */
 	for (;;)
 	{
+		CHECK_REQUEST;
+
 		/* do we need health checking for PostgreSQL? */
 		if (!degenerated && pool_config.health_check_period > 0)
 		{
@@ -341,11 +377,11 @@ int main(int argc, char **argv)
 			{
 				if (sts == -1)
 				{
-					failover_handler(SIGUSR1);		/* master down */
+					failover(SIGUSR1);		/* master down */
 				}
 				else if (sts == -2)
 				{
-					failover_handler(SIGUSR2);		/* secondary down */
+					failover(SIGUSR2);		/* secondary down */
 				}
 			}
 
@@ -356,14 +392,20 @@ int main(int argc, char **argv)
 			}
 
 			sleep_time = pool_config.health_check_period;
-			while (sleep_time > 0)
-			{
-				sleep_time = sleep(sleep_time);
-			}
+			pool_sleep(sleep_time);
 		}
 		else
 		{
-			pause();
+			for (;;)
+			{
+				int r;
+
+				POOL_SETMASK(&UnBlockSig);
+				r = pool_pause(NULL);
+				POOL_SETMASK(&BlockSig);
+				if (r > 0)
+					break;
+			}
 		}
 	}
 	return 0;
@@ -753,7 +795,6 @@ static RETSIGTYPE exit_handler(int sig)
 	myexit(0);
 }
 
-
 /*
  * handle SIGUSR1/SIGUSR2 (backend connection error, fail over request, if possible)
  *
@@ -761,12 +802,23 @@ static RETSIGTYPE exit_handler(int sig)
  * if sig == SIGUSR2, we assume that the secondary has been down.
  */
 static RETSIGTYPE failover_handler(int sig)
+{
+	POOL_SETMASK(&BlockSig);
+	failover_request = 1;
+	failover_signo = sig;
+	write(pipe_fds[1], "\0", 1);
+	POOL_SETMASK(&UnBlockSig);
+}
+
+/*
+ * Process failover request
+ * failover() must be called under protecting signals.
+ */
+static void failover(int sig)
 {
 	int i;
 	int replication = 0;
 
-	POOL_SETMASK(&BlockSig);
-
 	pool_debug("failover_handler called");
 
 	/*
@@ -776,7 +828,6 @@ static RETSIGTYPE failover_handler(int sig)
 	if (getpid() != mypid)
 	{
 		pool_debug("failover_handler: I am not parent");
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -785,7 +836,6 @@ static RETSIGTYPE failover_handler(int sig)
 	 */
 	if (exiting)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -794,7 +844,6 @@ static RETSIGTYPE failover_handler(int sig)
 	 */
 	if (switching)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -958,7 +1007,6 @@ static RETSIGTYPE failover_handler(int sig)
 
 		switching = 0;
 	}
-	POOL_SETMASK(&UnBlockSig);
 }
 
 /*
@@ -975,24 +1023,33 @@ static RETSIGTYPE health_check_timer_handler(int sig)
  * handle SIGCHLD
  */
 static RETSIGTYPE reap_handler(int sig)
+{
+	POOL_SETMASK(&BlockSig);
+	sigchld_request = 1;
+	write(pipe_fds[1], "\0", 1);
+	POOL_SETMASK(&UnBlockSig);
+}
+
+/*
+ * Attach zombie processes and restart child processes.
+ * reaper() must be called under protecting signals.
+ */
+static void reaper(void)
 {
 	pid_t pid;
 	int status;
 	int i;
 
-	POOL_SETMASK(&BlockSig);
-
 	pool_debug("reap_handler called");
+	sigchld_request = 0;
 
 	if (exiting)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
 	if (switching)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -1021,6 +1078,59 @@ static RETSIGTYPE reap_handler(int sig)
 				}
 			}
 		}
-		POOL_SETMASK(&UnBlockSig);
 
 	}
+
+/*
+ * pool_pause: A process pauses by select().
+ */
+static int pool_pause(struct timeval *timeout)
+{
+	fd_set rfds;
+	int n;
+	char dummy;
+
+	FD_ZERO(&rfds);
+	FD_SET(pipe_fds[0], &rfds);
+	n = select(pipe_fds[0]+1, &rfds, NULL, NULL, timeout);
+	if (n == 1)
+		read(pipe_fds[0], &dummy, 1);
+	return n;
+}
+
+/*
+ * pool_pause: A process sleep using pool_pause().
+ *             If a signal event occurs, it raises signal handler.
+ */
+static void pool_sleep(unsigned int second)
+{
+	struct timeval current_time, sleep_time;
+
+	gettimeofday(&current_time, NULL);
+	sleep_time.tv_sec = second + current_time.tv_sec;
+	sleep_time.tv_usec = current_time.tv_usec;
+
+	POOL_SETMASK(&UnBlockSig);
+	while (sleep_time.tv_sec > current_time.tv_sec ||
+		   sleep_time.tv_usec > current_time.tv_usec)
+	{
+		struct timeval timeout;
+		int r;
+
+		timeout.tv_sec = sleep_time.tv_sec - current_time.tv_sec;
+		timeout.tv_usec = sleep_time.tv_usec - current_time.tv_usec;
+		if (timeout.tv_usec < 0)
+		{
+			timeout.tv_sec--;
+			timeout.tv_usec += 1000000;
+		}
+
+		r = pool_pause(&timeout);
+		POOL_SETMASK(&BlockSig);
+		if (r > 0)									
+			CHECK_REQUEST;
+		POOL_SETMASK(&UnBlockSig);
+		gettimeofday(&current_time, NULL);
+	}
+	POOL_SETMASK(&BlockSig);
+}
