@@ -1,7 +1,7 @@
 /* -*-pgsql-c-*- */
 /*
  *
- * $Header: /cvsroot/pgpool/pgpool/pool.h,v 1.16 2007/02/03 04:22:21 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool.h,v 1.21 2007/07/26 08:45:42 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -60,11 +60,12 @@
 #define NO_STRICT_MODE(s) (strncasecmp((s), NO_STRICT_MODE_STR, strlen(NO_STRICT_MODE_STR)) == 0)
 
 typedef enum {
-    POOL_CONTINUE = 0,
+	POOL_CONTINUE = 0,
 	POOL_IDLE,
-    POOL_END,
-    POOL_ERROR,
-    POOL_FATAL
+	POOL_END,
+	POOL_ERROR,
+	POOL_FATAL,
+	POOL_DEADLOCK
 } POOL_STATUS;
 
 /* protocol major version numbers */
@@ -144,6 +145,7 @@ typedef struct {
 	int replication_stop_on_mismatch;		/* if there's a data mismatch between master and secondary
 											 * start degenration to stop replication mode
 											 */
+	int replicate_select; /* if non 0, replicate SELECT statement when load balancing is disabled. */
 	char **reset_query_list;		/* comma separated list of quries to be issued at the end of session */
 
 	int print_timestamp;		/* if non 0, print time stamp to each log line */
@@ -275,6 +277,7 @@ typedef struct {
  */
 extern POOL_CONFIG pool_config;	/* configuration values */
 extern POOL_CONNECTION_POOL *pool_connection_pool;	/* connection pool */
+extern volatile sig_atomic_t backend_timer_expired;
 extern long int weight_master;	/* normalized weight of master (0-RAND_MAX range) */
 extern char remote_ps_data[];		/* used for set_ps_display */
 
@@ -301,14 +304,16 @@ extern int pool_flush(POOL_CONNECTION *cp);
 extern int pool_flush_it(POOL_CONNECTION *cp);
 extern int pool_write_and_flush(POOL_CONNECTION *cp, void *buf, int len);
 extern char *pool_read_string(POOL_CONNECTION *cp, int *len, int line);
+extern int pool_unread(POOL_CONNECTION *cp, void *data, int len);
 
 extern int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 extern int pool_do_reauth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp);
 
 extern int pool_init_cp(void);
 extern POOL_CONNECTION_POOL *pool_create_cp(void);
-extern POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor);
+extern POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor, int check_socket);
 extern void pool_discard_cp(char *user, char *database, int protoMajor);
+extern void pool_backend_timer(void);
 
 extern POOL_STATUS ErrorResponse(POOL_CONNECTION *frontend, 
 								  POOL_CONNECTION_POOL *backend);
