@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/pool_connection_pool.c,v 1.7 2007/02/06 02:11:07 y-asaba Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool_connection_pool.c,v 1.12 2007/07/27 07:59:35 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -41,6 +41,7 @@
 #include "pool.h"
 
 POOL_CONNECTION_POOL *pool_connection_pool;	/* connection pool */
+volatile sig_atomic_t backend_timer_expired = 0;
 
 static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp, int secondary_backend);
 static POOL_CONNECTION_POOL *new_connection(POOL_CONNECTION_POOL *p);
@@ -65,7 +66,7 @@ int pool_init_cp(void)
 /*
 * find connection by user and database
 */
-POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor)
+POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor, int check_socket)
 {
 #ifdef HAVE_SIGPROCMASK
 	sigset_t oldmask;
@@ -97,17 +98,19 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor)
 			MASTER_CONNECTION(p)->closetime = 0;
 			POOL_SETMASK(&oldmask);
 
-			if (check_socket_status(MASTER(p)->fd) < 0 ||
-				(DUAL_MODE && check_socket_status(MASTER(p)->fd) < 0))
+			if (check_socket &&
+				(check_socket_status(MASTER(p)->fd) < 0 ||
+				 (DUAL_MODE && check_socket_status(MASTER(p)->fd) < 0)))
 			{
 				pool_log("connection closed. retry to create new connection pool.");
 				pool_free_startup_packet(MASTER_CONNECTION(p)->sp);
 				pool_close(MASTER_CONNECTION(p)->con);
+				free(MASTER_CONNECTION(p));
 
 				if (DUAL_MODE)
 				{
-					/* do not free memory! we did not allocate them */
 					pool_close(SECONDARY_CONNECTION(p)->con);
+					free(SECONDARY_CONNECTION(p));
 				}
 
 				memset(p, 0, sizeof(POOL_CONNECTION_POOL));
@@ -128,7 +131,7 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor)
  */
 void pool_discard_cp(char *user, char *database, int protoMajor)
 {
-	POOL_CONNECTION_POOL *p = pool_get_cp(user, database, protoMajor);
+	POOL_CONNECTION_POOL *p = pool_get_cp(user, database, protoMajor, 0);
 
 	if (p == NULL)
 	{
@@ -138,11 +141,13 @@ void pool_discard_cp(char *user, char *database, int protoMajor)
 
 	pool_free_startup_packet(MASTER_CONNECTION(p)->sp);
 	pool_close(MASTER_CONNECTION(p)->con);
+	free(MASTER_CONNECTION(p));
 
 	if (DUAL_MODE)
 	{
 		/* do not free memory! we did not allocate them */
 		pool_close(SECONDARY_CONNECTION(p)->con);
+		free(SECONDARY_CONNECTION(p));
 	}
 
 	memset(p, 0, sizeof(POOL_CONNECTION_POOL));
@@ -204,11 +209,13 @@ POOL_CONNECTION_POOL *pool_create_cp(void)
 
 	pool_free_startup_packet(MASTER_CONNECTION(p)->sp);
 	pool_close(MASTER_CONNECTION(p)->con);
+	free(MASTER_CONNECTION(p));
 
 	if (DUAL_MODE)
 	{
 		/* do not free memory! we did not allocate them */
 		pool_close(SECONDARY_CONNECTION(p)->con);
+		free(SECONDARY_CONNECTION(p));
 	}
 
 	memset(p, 0, sizeof(POOL_CONNECTION_POOL));
@@ -253,6 +260,11 @@ void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend)
  * backend connection close timer handler
  */
 RETSIGTYPE pool_backend_timer_handler(int sig)
+{
+	backend_timer_expired = 1;
+}
+
+void pool_backend_timer(void)
 {
 #define TMINTMAX 0x7fffffff
 
@@ -289,11 +301,12 @@ RETSIGTYPE pool_backend_timer_handler(int sig)
 
 				pool_free_startup_packet(MASTER_CONNECTION(p)->sp);
 				pool_close(MASTER_CONNECTION(p)->con);
+				free(MASTER_CONNECTION(p));
 
 				if (DUAL_MODE)
 				{
-					/* do not free memory! we did not allocate them */
 					pool_close(SECONDARY_CONNECTION(p)->con);
+					free(SECONDARY_CONNECTION(p));
 				}
 
 				memset(p, 0, sizeof(POOL_CONNECTION_POOL));
@@ -367,12 +380,20 @@ int connect_inet_domain_socket(int secondary_backend)
 			(char *) hp->h_addr,
 			hp->h_length);
 
-	if (connect(fd, (struct sockaddr *)&addr, len) < 0)
+	for (;;)
 	{
-		pool_error("connect_inet_domain_socket: connect() failed: %s",strerror(errno));
-		close(fd);
-		return -1;
+		if (connect(fd, (struct sockaddr *)&addr, len) < 0)
+		{
+			if (errno == EINTR || errno == EAGAIN)
+				continue;
+
+			pool_error("connect_inet_domain_socket: connect() failed: %s",strerror(errno));
+			close(fd);
+			return -1;
+		}
+		break;
 	}
+
 	return fd;
 }
 
@@ -400,11 +421,18 @@ int connect_unix_domain_socket(int secondary_backend)
 
 	len = sizeof(struct sockaddr_un);
 
-	if (connect(fd, (struct sockaddr *)&addr, len) < 0)
+	for (;;)
 	{
-		pool_error("connect_unix_domain_socket: connect() failed: %s", strerror(errno));
-		close(fd);
-		return -1;
+		if (connect(fd, (struct sockaddr *)&addr, len) < 0)
+		{
+			if (errno == EINTR || errno == EAGAIN )
+				continue;
+
+			pool_error("connect_unix_domain_socket: connect() failed: %s", strerror(errno));
+			close(fd);
+			return -1;
+		}
+		break;
 	}
 	pool_debug("connected to postmaster Unix domain socket: %s fd: %d", addr.sun_path, fd);
 	return fd;
