@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
-* $Header: /cvsroot/pgpool/pgpool/pool_stream.c,v 1.4 2005/07/30 12:30:20 t-ishii Exp $
+* $Header: /cvsroot/pgpool/pgpool/pool_stream.c,v 1.10 2007/07/11 01:40:17 y-asaba Exp $
 *
 * pgpool: a language independent connection pool server for PostgreSQL 
 * written by Tatsuo Ishii
 *
-* Copyright (c) 2003-2005	Tatsuo Ishii
+* Copyright (c) 2003-2007	PgPool Global Development Group
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
@@ -83,6 +83,12 @@ POOL_CONNECTION *pool_open(int fd)
 	cp->sbufsz = 0;
 
 	cp->fd = fd;
+
+	memset((void *) &cp->raddr, 0, sizeof(cp->raddr));
+	cp->auth_arg = NULL;
+	cp->database = NULL;
+	cp->username = NULL;
+
 	return cp;
 }
 
@@ -98,7 +104,14 @@ void pool_close(POOL_CONNECTION *cp)
 		free(cp->sbuf);
 	if (cp->buf2)
 		free(cp->buf2);
-	pool_discard_params(&cp->params);
+	if (cp->isbackend)
+		pool_discard_params(&cp->params);
+	if (cp->auth_arg)
+		free(cp->auth_arg);
+	if (cp->database)
+		free(cp->database);
+	if (cp->username)
+		free(cp->username);
 	free(cp);
 }
 
@@ -751,3 +764,31 @@ static int consume_pending_data(POOL_CONNECTION *cp, void *data, int len)
 
 	return consume_size;
 }
+
+/*
+ * pool_unread: Put back data to input buffer
+ */
+int pool_unread(POOL_CONNECTION *cp, void *data, int len)
+{
+	void *p = cp->hp;
+	int n = cp->len + len;
+	int realloc_size;
+	
+	if (cp->bufsz < n)
+	{
+		realloc_size = (n/READBUFSZ+1)*READBUFSZ;
+		p = realloc(cp->hp, realloc_size);
+		if (p == NULL)
+		{
+			pool_error("pool_unread: realloc failed");
+			return -1;
+		}
+		cp->hp = p;
+	}
+	if (cp->len != 0)
+		memmove(p + len, cp->hp + cp->po, cp->len);
+	memmove(p, data, len);
+	cp->len = n;
+	cp->po = 0;
+	return 0;
+}
